
#ifdef _MSC_VER
  #include "genio.h"
  #include "hpgcc49.h"
  #include "hpgraphics.h"
  #define main hp_main
#endif

#ifdef BUCHAS_C
  #include "genio.h"
  #include <hpgcc49.h>
  #include <hpgraphics.h>
  #define main hp_main
#endif

#if !defined(BUCHAS_C) && !defined(_MSC_VER)
  #include <hpgcc49.h>
  #include <hpgraphics.h>
#endif

// This is the "XorShift" fast pseudorandom number generator

typedef unsigned int uint32_t;

uint32_t xor128(void) 
{
  static uint32_t x = 123456789;
  static uint32_t y = 362436069;
  static uint32_t z = 521288629;
  static uint32_t w = 88675123;

  uint32_t t;
 
  t = x ^ (x << 11);
  x = y; y = z; z = w;
  return w = w ^ (w >> 19) ^ (t ^ (t >> 8));
}

#define POWER 6
//#define DIAG

#define HEIGHT (1 << POWER)
#define WIDTH  HEIGHT

#define X_OFFSET 5
#define Y_OFFSET 5

char count[HEIGHT * WIDTH];
char state[HEIGHT * WIDTH];
char newcount[HEIGHT * WIDTH];

// UPDATE LIFE COUNT FOR NEIGHBORS

void add_neighbors(int j, int i, int value)
{
  int offset0;
  int leftadd, rightadd, topadd, botadd;

  if(i == 0) 
    leftadd = WIDTH - 1; 
  else 
    leftadd = -1;
  
  if(i == WIDTH - 1) 
    rightadd = -WIDTH + 1; 
  else 
    rightadd = 1;

  if(j == 0) 
    topadd = (HEIGHT - 1) * WIDTH; 
  else 
    topadd = -WIDTH;

  if(j == HEIGHT - 1) 
    botadd = -(HEIGHT - 1) * WIDTH; 
  else 
    botadd = WIDTH;

  offset0 = j * WIDTH + i;

  // TOP-LEFT
  #ifdef DIAG
  newcount[offset0 + topadd + leftadd ] += value;
  #endif
  // TOP
  newcount[offset0 + topadd           ] += value;
  // TOP-RIGHT
  #ifdef DIAG
  newcount[offset0 + topadd + rightadd] += value;
  #endif
  // LEFT
  newcount[offset0 + leftadd          ] += value;
  // RIGHT
  newcount[offset0 + rightadd         ] += value;
  // BOT-LEFT
  #ifdef DIAG
  newcount[offset0 + botadd + leftadd ] += value;
  #endif
  // BOT
  newcount[offset0 + botadd           ] += value;
  // BOT-RIGHT
  #ifdef DIAG
  newcount[offset0 + botadd + rightadd] += value;
  #endif
}


// DETERMINE NEW STATE AFTER ONE TICK

void play_tick(void)
{
  int i, j;
  int offset;
  
  // COPY COUNT TO TEMPORARY STORAGE
  memmove(newcount, count, sizeof(count));

  offset = 0;

  for(j = 0; j < HEIGHT; ++j)
  {
    for(i = 0; i < WIDTH; ++i)
    {
      if(state[offset]) 
      {
        // IF ALIVE, CHECK IF IT SHOULD DIE
        if((count[offset] | 1) != 3 ) 
        {
          // KILL IT
          state[offset] = 0;
          hpg_set_color(hpg_stdscreen, HPG_COLOR_WHITE);
          hpg_draw_pixel(i + X_OFFSET, j + Y_OFFSET);
          // UPDATE COUNT FOR ALL NEIGHBORS
          add_neighbors(j, i, -1);
        }
      }
      else 
      {
        // IF DEAD, CHECK IF IT SHOULD COME TO LIFE
        if(count[offset] == 3) 
        {
          state[offset] = 1;
          hpg_set_color(hpg_stdscreen, HPG_COLOR_BLACK);
          hpg_draw_pixel(i + X_OFFSET, j + Y_OFFSET);
          // UPDATE NEIGHBORS COUNT
          add_neighbors(j, i, 1);
        }
      }
      offset++;
    }
  }

  // MOVE BACK TO PERMANENT STORAGE
  memmove(count, newcount, sizeof(count));
}

// SET INITIAL STATE
void calc_count(void)
{
  int i, j;
  int offset;

  // COPY COUNT TO TEMPORARY STORAGE
  memset((void *)newcount, 0, sizeof(newcount));
  
  offset = 0;
  
  for(j = 0; j < HEIGHT; ++j)
  {
    for(i = 0; i < WIDTH; ++i)
    {
      if(state[offset]) 
      { 
        add_neighbors(j, i, 1); 
        // hpg_set_color(hpg_stdscreen, HPG_COLOR_BLACK); 
      }
      else 
      { 
        // hpg_set_color(hpg_stdscreen, HPG_COLOR_WHITE); 
      } 
      // hpg_draw_pixel(i + X_OFFSET, j + Y_OFFSET);
      offset++;
    }
  }
  // MOVE BACK TO PERMANENT STORAGE
  memmove(count, newcount, sizeof(count));
}

static void fillr(int x, int y, int w, int h, int color)
{
  int i, j;
  hpg_set_color(hpg_stdscreen, (unsigned char)color);
  for(j = 0; j < h; j++)
    for(i = 0; i < w; i++)
      hpg_draw_pixel(x + i, y + j);
}

static void ising_step(void);

int main(void)
{
  int i, j, offset;

  hpg_clear();
  
  // SET TO RANDOM STATE

  fillr(X_OFFSET - 1, Y_OFFSET - 1,      WIDTH + 2, 1, HPG_COLOR_BLACK);
  fillr(X_OFFSET - 1, Y_OFFSET + HEIGHT, WIDTH + 2, 1, HPG_COLOR_BLACK);
  fillr(X_OFFSET - 1,     Y_OFFSET, 1, HEIGHT, HPG_COLOR_BLACK);
  fillr(X_OFFSET + WIDTH, Y_OFFSET, 1, HEIGHT, HPG_COLOR_BLACK);
  for(j = 0; j < HEIGHT; j++)
  {
    for(i = 0; i < WIDTH; i++)
    {
      if((xor128() & 0x80000000UL) != 0)
        hpg_set_color(hpg_stdscreen, HPG_COLOR_BLACK); 
      else
        hpg_set_color(hpg_stdscreen, HPG_COLOR_WHITE); 
      hpg_draw_pixel(i + X_OFFSET, j + Y_OFFSET);
    }
  }

  // GET STATE MATRIX
  offset = 0;
  for(j = 0; j < HEIGHT; j++)
  {
    for(i = 0; i < WIDTH; i++)
    {
      if(hpg_get_pixel(hpg_stdscreen, i + X_OFFSET, j + Y_OFFSET)) 
        state[offset] = 1;
      else 
        state[offset] = 0;
      offset++;
    }
  }
  
  // INITIALISE ITERATOR
  calc_count();
  
  while(!keyb_isON())
  {
    // play_tick();
    ising_step();
  }
  return 0;
}

static void ising_step(void)
{
  unsigned i, j, offset;
  uint32_t rnd1;
  
  unsigned char old;

  rnd1 = xor128();

  i = rnd1 & (HEIGHT - 1);
  rnd1 >>= POWER;
  j = rnd1 & (WIDTH - 1);
  rnd1 >>= POWER;

  rnd1 &= ((1UL << 16) - 1);

  offset = (i << POWER) + j;
  old = state[offset];

  //if(rnd1 & 1)
  {
    int update;

    if(old != 0)
    {
      hpg_set_color(hpg_stdscreen, HPG_COLOR_WHITE); 
      update = -1;
    }
    else
    {
      hpg_set_color(hpg_stdscreen, HPG_COLOR_BLACK); 
      update = 1;
    }
    hpg_draw_pixel(i + X_OFFSET, j + Y_OFFSET);
    add_neighbors(i, j, update);
    state[offset] = 1 - old;
  }

}
